import asyncio
from collections import defaultdict
from dataclasses import dataclass, field
from typing import ClassVar, DefaultDict, Dict, Optional, Tuple

from kstreams import ConsumerRecord

from . import test_clients


@dataclass
class Topic:
    name: str
    queue: asyncio.Queue
    total_partition_events: DefaultDict[int, int] = field(
        default_factory=lambda: defaultdict(lambda: -1)
    )
    total_events: int = 0
    # for now we assumed that 1 streams is connected to 1 topic
    consumer: Optional["test_clients.Consumer"] = None

    async def put(self, event: ConsumerRecord) -> None:
        await self.queue.put(event)
        self._inc_amount(event)

    async def get(self) -> ConsumerRecord:
        cr = await self.queue.get()
        self.task_done()
        return cr

    def get_nowait(self) -> ConsumerRecord:
        return self.queue.get_nowait()

    def put_nowait(self, *, event: ConsumerRecord) -> None:
        self.queue.put_nowait(event)
        self._inc_amount(event)

    def _inc_amount(self, event: ConsumerRecord) -> None:
        # keep track of the amount of events per topic partition
        self.total_partition_events[event.partition] += 1
        self.total_events += 1

    def task_done(self) -> None:
        self.queue.task_done()

    async def join(self) -> None:
        await self.queue.join()

    def is_empty(self) -> bool:
        return self.queue.empty()

    def size(self) -> int:
        return self.queue.qsize()

    def offset(self, *, partition: int) -> int:
        return self.total_partition_events[partition]

    @property
    def consumed(self) -> bool:
        """
        We need to check if the Topic has a Consumer and if there are messages in it
        """
        return self.consumer is None or self.is_empty()

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return (
            f"Topic {self.name} with Consumer: {self.consumer}. "
            f"Messages in buffer: {self.queue.qsize()}"
        )


@dataclass
class TopicManager:
    # The queues will represent the kafka topics during testing
    # where the name is the topic name
    topics: ClassVar[Dict[str, Topic]] = {}

    @classmethod
    def get(cls, name: str) -> Topic:
        topic = cls.topics.get(name)

        if topic is not None:
            return topic
        raise ValueError(
            f"You might be trying to get the topic {name} outside the "
            "`client async context` or trying to get an event from an empty "
            f"topic {name}. Make sure that the code is inside the async context"
            "and the topic has events."
        )

    @classmethod
    def create(
        cls, name: str, consumer: Optional["test_clients.Consumer"] = None
    ) -> Topic:
        topic = Topic(name=name, queue=asyncio.Queue(), consumer=consumer)
        cls.topics[name] = topic
        return topic

    @classmethod
    def get_or_create(
        cls, name: str, consumer: Optional["test_clients.Consumer"] = None
    ) -> Tuple[Topic, bool]:
        """
        A convenience method for looking up Topic by name.
        If the topic does not exist a new one is created.

        Returns a tuple of (Topic, created), where Topic is the
        retrieved or created object and created is a boolean
        specifying whether a new Topic was created.
        """
        try:
            topic = cls.get(name)
            return topic, False
        except ValueError:
            topic = cls.create(name, consumer=consumer)
            return topic, True

    @classmethod
    def all_messages_consumed(cls) -> bool:
        """
        Check if all the messages has been consumed for ALL the topics
        """
        for topic in cls.topics.values():
            # if there is at least 1 topic with events we need to keep waiting
            if not topic.consumed:
                return False
        return True

    @classmethod
    async def join(cls) -> None:
        """
        Wait for all topic messages to be processed.
        Only topics that have a consumer assigned should be awaited.
        """
        await asyncio.gather(
            *[topic.join() for topic in cls.topics.values() if not topic.consumed]
        )

    @classmethod
    def clean(cls) -> None:
        cls.topics = {}



from types import TracebackType
from typing import Any, Dict, List, Optional, Type

from kstreams import ConsumerRecord
from kstreams.engine import StreamEngine
from kstreams.prometheus.monitor import PrometheusMonitor
from kstreams.serializers import Serializer
from kstreams.streams import Stream
from kstreams.types import Headers

from .structs import RecordMetadata
from .test_clients import TestConsumer, TestProducer
from .topics import Topic, TopicManager


class TestMonitor(PrometheusMonitor):
    __test__ = False

    def start(self, *args, **kwargs) -> None:
        print("herte....")
        # ...

    async def stop(self, *args, **kwargs) -> None:
        ...

    def add_topic_partition_offset(self, *args, **kwargs) -> None:
        ...

    def clean_stream_consumer_metrics(self, *args, **kwargs) -> None:
        ...

    def add_producer(self, *args, **kwargs):
        ...

    def add_streams(self, *args, **kwargs):
        ...


class TestStreamClient:
    __test__ = False

    def __init__(
        self,
        stream_engine: StreamEngine,
        monitoring_enabled: bool = True,
        topics: Optional[List[str]] = None,
    ) -> None:
        self.stream_engine = stream_engine

        # Extra topics' names defined by the end user which must be created
        # before the cycle test starts
        self.extra_user_topics = topics

        # store the user clients to restore them later
        self.monitor = stream_engine.monitor
        self.producer_class = self.stream_engine.producer_class
        self.consumer_class = self.stream_engine.consumer_class

        self.stream_engine.producer_class = TestProducer
        self.stream_engine.consumer_class = TestConsumer

        if not monitoring_enabled:
            self.stream_engine.monitor = TestMonitor()

    def mock_streams(self) -> None:
        streams: List[Stream] = self.stream_engine._streams
        for stream in streams:
            stream.consumer_class = TestConsumer

    def setup_mocks(self) -> None:
        self.mock_streams()

    def create_extra_topics(self) -> None:
        if self.extra_user_topics is not None:
            for topic_name in self.extra_user_topics:
                TopicManager.create(topic_name)

    async def start(self) -> None:
        self.setup_mocks()
        self.create_extra_topics()
        await self.stream_engine.start()

    async def stop(self) -> None:
        # If there are streams, we must wait until all the messages are consumed
        if self.stream_engine._streams:
            await TopicManager.join()
        await self.stream_engine.stop()

        # restore original config
        self.stream_engine.producer_class = self.producer_class
        self.stream_engine.consumer_class = self.consumer_class
        self.stream_engine.monitor = self.monitor

        # clean the topics after finishing the test to make sure that
        # no data is left tover
        TopicManager.clean()

    async def __aenter__(self) -> "TestStreamClient":
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_t: Optional[Type[BaseException]],
        exc_v: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.stop()

    async def send(
        self,
        topic: str,
        value: Any = None,
        key: Optional[Any] = None,
        partition: int = 0,
        timestamp_ms: Optional[int] = None,
        headers: Optional[Headers] = None,
        serializer: Optional[Serializer] = None,
        serializer_kwargs: Optional[Dict] = None,
    ) -> RecordMetadata:
        return await self.stream_engine.send(
            topic,
            value=value,
            key=key,
            partition=partition,
            timestamp_ms=timestamp_ms,
            headers=headers,
            serializer=serializer,
            serializer_kwargs=serializer_kwargs,
        )

    def get_topic(self, *, topic_name: str) -> Topic:
        return TopicManager.get(topic_name)

    async def get_event(self, *, topic_name: str) -> ConsumerRecord:
        topic = TopicManager.get(topic_name)
        return await topic.get()



import contextlib
import ssl
from tempfile import NamedTemporaryFile
from typing import Optional, Union

from aiokafka.helpers import create_ssl_context as aiokafka_create_ssl_context

from kstreams import types


def encode_headers(headers: types.Headers) -> types.EncodedHeaders:
    return [(header, value.encode()) for header, value in headers.items()]


def create_ssl_context_from_mem(
    *,
    certdata: str,
    keydata: str,
    password: Optional[str] = None,
    cadata: Optional[str] = None,
) -> Optional[ssl.SSLContext]:
    """Create a SSL context from data on memory.

    This makes it easy to read the certificates from environmental variables
    Usually the data is loaded from env variables.

    Arguments:
        cadata: certificates used to sign broker certificates provided as unicode str
        certdata: the client certificate, as well as any CA certificates needed to
            establish the certificate's authenticity provided as unicode str
        keydata: the client private key provided as unicode str
        password: optional password to be used when loading the
            certificate chain
    """
    with contextlib.ExitStack() as stack:
        cert_file = stack.enter_context(NamedTemporaryFile(suffix=".crt"))
        key_file = stack.enter_context(NamedTemporaryFile(suffix=".key"))

        # expecting unicode data, writing it as bytes to files as utf-8
        cert_file.write(certdata.encode("utf-8"))
        cert_file.flush()

        key_file.write(keydata.encode("utf-8"))
        key_file.flush()

        ssl_context = ssl.create_default_context(cadata=cadata)
        ssl_context.load_cert_chain(
            cert_file.name, keyfile=key_file.name, password=password
        )
        return ssl_context
    return None


def create_ssl_context(
    *,
    cafile: str = None,
    capath: str = None,
    cadata: Union[str, bytes] = None,
    certfile: str = None,
    keyfile: str = None,
    password: str = None,
    crlfile=None,
):
    """Wrapper of [aiokafka.helpers.create_ssl_context](
        https://aiokafka.readthedocs.io/en/stable/api.html#helpers
    )
    with typehints.

    Arguments:
        cafile: Certificate Authority file path containing certificates
            used to sign broker certificates
        capath: Same as `cafile`, but points to a directory containing
            several CA certificates
        cadata: Same as `cafile`, but instead contains already
            read data in either ASCII or bytes format
        certfile: optional filename of file in PEM format containing
            the client certificate, as well as any CA certificates needed to
            establish the certificate's authenticity
        keyfile: optional filename containing the client private key.
        password: optional password to be used when loading the
            certificate chain

    """
    return aiokafka_create_ssl_context(
        cafile=cafile,
        capath=capath,
        cadata=cadata,
        certfile=certfile,
        keyfile=keyfile,
        password=password,
        crlfile=crlfile,
    )


__all__ = ["create_ssl_context", "create_ssl_context_from_mem", "encode_headers"]



import typing

from kstreams import ConsumerRecord, RecordMetadata

if typing.TYPE_CHECKING:
    from .serializers import Serializer  #  pragma: no cover

Headers = typing.Dict[str, str]
EncodedHeaders = typing.Sequence[typing.Tuple[str, bytes]]
StreamFunc = typing.Callable
NextMiddlewareCall = typing.Callable[[ConsumerRecord], typing.Awaitable[None]]


class Send(typing.Protocol):
    def __call__(
        self,
        topic: str,
        value: typing.Any = None,
        key: typing.Any = None,
        partition: typing.Optional[int] = None,
        timestamp_ms: typing.Optional[int] = None,
        headers: typing.Optional[Headers] = None,
        serializer: typing.Optional["Serializer"] = None,
        serializer_kwargs: typing.Optional[typing.Dict] = None,
    ) -> typing.Awaitable[RecordMetadata]:
        ...



from typing import Optional, Type

from .backends.kafka import Kafka
from .clients import Consumer, ConsumerType, Producer, ProducerType
from .engine import StreamEngine
from .prometheus.monitor import PrometheusMonitor
from .serializers import Deserializer, Serializer


def create_engine(
    title: Optional[str] = None,
    backend: Optional[Kafka] = None,
    consumer_class: Type[ConsumerType] = Consumer,
    producer_class: Type[ProducerType] = Producer,
    serializer: Optional[Serializer] = None,
    deserializer: Optional[Deserializer] = None,
    monitor: Optional[PrometheusMonitor] = None,
) -> StreamEngine:
    if monitor is None:
        monitor = PrometheusMonitor()

    if backend is None:
        backend = Kafka()

    return StreamEngine(
        backend=backend,
        title=title,
        consumer_class=consumer_class,
        producer_class=producer_class,
        serializer=serializer,
        deserializer=deserializer,
        monitor=monitor,
    )



APPLICATION_X_AVRO_BINARY = "application/x-avro-binary"
APPLICATION_X_AVRO_JSON = "application/x-avro-json"
APPLICATION_X_JSON_SCHEMA = "application/x-json-schema"
APPLICATION_JSON = "application/json"

AVRO_SERIALIZATION_TYPES = (
    APPLICATION_X_AVRO_BINARY,
    APPLICATION_X_AVRO_JSON,
)
SERIALIZATION_TYPES = AVRO_SERIALIZATION_TYPES + (
    APPLICATION_X_JSON_SCHEMA,
    APPLICATION_JSON,
)



import enum
import inspect
from typing import Any, Callable


class UDFType(str, enum.Enum):
    NO_TYPING = "NO_TYPING"
    CR_ONLY_TYPING = "CR_ONLY_TYPING"
    ALL_TYPING = "ALL_TYPING"


def inspect_udf(func: Callable, a_type: Any) -> UDFType:
    """
    Inspect the user defined function (coroutine) to get the  proper way to call it

    The cases are:

        1. Using only the Stream with ot without typing. This is the classical way
            to use kstreams. This might be deprecated.

            Note: that the `argument` streams, can be anything like consumer,
            stream, processor, etc.

            @stream_engine.stream(topic, name="my-stream")
                async def consume(stream: Stream):
                    for cr in stream:
                        ...

            or

            @stream_engine.stream(topic, name="my-stream")
                async def consume(stream):
                    for cr in stream:
                        ...

        2. Using only the ConsumerRecord (it must be used with typing)

            @stream_engine.stream(topic, name="my-stream")
                async def consume(cr: ConsumerRecord):
                    ...

        3. Using ConsumerRecord and Stream, both with typing.
            The order is important as they are arguments and not kwargs

            @stream_engine.stream(topic, name="my-stream")
                async def consume(cr: ConsumerRecord, stream: Stream):
                    ...
    """
    signature = inspect.signature(func)

    # order is important as there are not kwargs
    parameters = list(signature.parameters.values())

    if len(parameters) == 1:
        # use case 1 or 2
        first_parameter = parameters[0]
        if first_parameter.annotation in (inspect._empty, a_type):
            # use case 1
            return UDFType.NO_TYPING
        else:
            # use case 2
            return UDFType.CR_ONLY_TYPING
    else:
        # use case 3
        return UDFType.ALL_TYPING



from datetime import datetime
from typing import Any, Coroutine, Dict, List, Optional, Set, Tuple

from kstreams import ConsumerRecord, RebalanceListener, TopicPartition
from kstreams.clients import Consumer, Producer
from kstreams.serializers import Serializer
from kstreams.types import Headers

from .structs import RecordMetadata
from .topics import TopicManager


class Base:
    async def start(self):
        ...


class TestProducer(Base, Producer):
    __test__ = False

    async def send(
        self,
        topic_name: str,
        value: Any = None,
        key: Any = None,
        partition: int = 0,
        timestamp_ms: Optional[float] = None,
        headers: Optional[Headers] = None,
        serializer: Optional[Serializer] = None,
        serializer_kwargs: Optional[Dict] = None,
    ) -> Coroutine:
        topic, _ = TopicManager.get_or_create(topic_name)
        timestamp_ms = timestamp_ms or datetime.now().timestamp()
        total_partition_events = topic.offset(partition=partition)
        partition = partition or 0

        consumer_record = ConsumerRecord(
            topic=topic_name,
            value=value,
            key=key,
            headers=headers,
            partition=partition,
            timestamp=timestamp_ms,
            offset=total_partition_events + 1,
            timestamp_type=None,
            checksum=None,
            serialized_key_size=None,
            serialized_value_size=None,
        )

        await topic.put(consumer_record)

        async def fut():
            return RecordMetadata(
                topic=topic_name,
                partition=partition,
                timestamp=timestamp_ms,
                offset=total_partition_events + 1,
            )

        return fut()


class TestConsumer(Base, Consumer):
    __test__ = False

    def __init__(self, group_id: Optional[str] = None, **kwargs) -> None:
        # copy the aiokafka behavior
        self.topics: Optional[Tuple[str]] = None
        self._group_id: Optional[str] = group_id
        self._assignment: List[TopicPartition] = []
        self.partitions_committed: Dict[TopicPartition, int] = {}

        # Called to make sure that has all the kafka attributes like _coordinator
        # so it will behave like an real Kafka Consumer
        super().__init__()

    def subscribe(
        self,
        *,
        topics: Tuple[str],
        listener: RebalanceListener,
        **kwargs,
    ) -> None:
        self.topics = topics

        for topic_name in topics:
            topic, created = TopicManager.get_or_create(topic_name, consumer=self)

            if not created:
                # It means that the topic already exist, so we are in
                # the situation where the topic hs events and the Stream
                # was added on runtime
                topic.consumer = self

            for partition_number in range(0, 3):
                self._assignment.append(
                    TopicPartition(topic=topic_name, partition=partition_number)
                )

        if listener.stream is not None:
            listener.stream.seek_to_initial_offsets()

    def assignment(self) -> List[TopicPartition]:
        return self._assignment

    def _check_partition_assignments(self, consumer_record: ConsumerRecord) -> None:
        """
        When an event is consumed the partition can be any positive int number
        because there is not limit in the producer side (only during testing of course).
        In case that the partition is not in the `_assignment` we need to register it.

        This is only during testing as in real use cases the assignments happens
        at the moment of kafka bootstrapping
        """
        topic_partition = TopicPartition(
            topic=consumer_record.topic,
            partition=consumer_record.partition,
        )

        if topic_partition not in self._assignment:
            self._assignment.append(topic_partition)

    def last_stable_offset(self, topic_partition: TopicPartition) -> int:
        topic = TopicManager.get(topic_partition.topic)

        if topic is not None:
            return topic.offset(partition=topic_partition.partition)
        return -1

    async def position(self, topic_partition: TopicPartition) -> int:
        """
        Get the offset of the *next record* that will be fetched,
        so it returns offset(topic_partition) + 1
        """
        return self.last_stable_offset(topic_partition) + 1

    def highwater(self, topic_partition: TopicPartition) -> int:
        """
        A highwater offset is the offset that will be assigned to
        the *next message* that is produced, so it returns
        offset(topic_partition) + 1
        """
        return self.last_stable_offset(topic_partition) + 1

    async def commit(self, offsets: Optional[Dict[TopicPartition, int]] = None) -> None:
        if offsets is not None:
            for topic_partition, offset in offsets.items():
                self.partitions_committed[topic_partition] = offset
        return None

    async def committed(self, topic_partition: TopicPartition) -> Optional[int]:
        return self.partitions_committed.get(topic_partition, 0)

    async def end_offsets(
        self, partitions: List[TopicPartition]
    ) -> Dict[TopicPartition, int]:
        topic = TopicManager.get(partitions[0].topic)
        end_offsets = {
            topic_partition: topic.offset(partition=topic_partition.partition) + 1
            for topic_partition in partitions
        }
        return end_offsets

    def partitions_for_topic(self, topic: str) -> Set:
        """
        Return the partitions of all assigned topics. The `topic` argument is not used
        because in a testing enviroment the only topics are the ones declared by the end
        user.

        The AIOKafkaConsumer returns a Set, so we do the same.
        """
        partitions = [topic_partition.partition for topic_partition in self._assignment]
        return set(partitions)

    async def getone(
        self,
    ) -> Optional[ConsumerRecord]:  # The return type must be fixed later on
        topic = None
        for topic_partition in self._assignment:
            topic = TopicManager.get(topic_partition.topic)

            if not topic.consumed:
                break

        if topic is not None:
            consumer_record = await topic.get()
            self._check_partition_assignments(consumer_record)
            return consumer_record

        return None

    async def getmany(
        self,
        *partitions: List[TopicPartition],
        timeout_ms: int = 0,
        max_records: int = 1,
    ) -> Dict[TopicPartition, List[ConsumerRecord]]:
        """
        Basic getmany implementation.
        `partitions` and `timeout_ms` could be added to the logic
        but it seems unnecessary for now; if end users request them we
        can add it
        """
        return {
            self._assignment[0]: [await self.getone() for _ in range(0, max_records)]
        }

    def seek(self, *, partition: TopicPartition, offset: int) -> None:
        # This method intends to have the same signature as aiokafka but with kwargs
        # rather than positional arguments
        topics = self.topics or ()

        if partition.topic in topics:
            topic = TopicManager.get(name=partition.topic)
            partition_offset = topic.offset(partition=partition.partition)

            # only consume if the offset to seek if <= the parition total events
            if offset <= partition_offset:
                consumed_events = 0

                # keep consuming if the events to consume <= offset to seek
                while consumed_events < offset:
                    event = topic.get_nowait()
                    topic.task_done()

                    if event.partition == partition.partition:
                        # only decrease if the event.partition matches
                        # the partition that the user wants to seek
                        consumed_events += 1
                    else:
                        # ideally each partition should be a Queue
                        # for now just add the same event to the queue
                        topic.put_nowait(event=event)

                    # it means that this consumer can consume
                    # from the TopicPartition so we can add it
                    # to the _assignment
                    if partition not in self._assignment:
                        self._assignment.append(partition)



import ssl
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, ConfigDict, model_validator


class SecurityProtocol(str, Enum):
    SSL = "SSL"
    PLAINTEXT = "PLAINTEXT"
    SASL_PLAINTEXT = "SASL_PLAINTEXT"
    SASL_SSL = "SASL_SSL"


class SaslMechanism(str, Enum):
    PLAIN = "PLAIN"
    GSSAPI = "GSSAPI"
    SCRAM_SHA_256 = "SCRAM-SHA-256"
    SCRAM_SHA_512 = "SCRAM-SHA-512"
    OAUTHBEARER = "OAUTHBEARER"


class Kafka(BaseModel):
    """
    The `Kafka` backend validates the given attributes.

    It uses pydantic internally.

    !!! Example
        ```python
        from kstreams.backends.kafka import Kafka
        from kstreams import create_engine, Stream

        backend = Kafka(bootstrap_servers=["localhost:8082"])
        stream_engine = create_engine(title="my-stream-engine", backend=backend)
        ```

    Attributes:
        bootstrap_servers: kafka list of `hostname:port`
        security_protocol: Protocol used to communicate with brokers
        ssl_context: a python std `ssl.SSLContext` instance, you can generate
            it with `create_ssl_context`
            or `create_ssl_context_from_mem`
        sasl_mechanism: Authentication mechanism when `security_protocol` is configured
            for `SASL_PLAINTEXT` or `SASL_SSL`
        sasl_plain_username: username for sasl PLAIN authentication
        sasl_plain_password: password for sasl PLAIN authentication
        sasl_oauth_token_provider: smth

    Raises:
        ValidationError: a `pydantic.ValidationError` exception
    """

    bootstrap_servers: List[str] = ["localhost:9092"]
    security_protocol: SecurityProtocol = SecurityProtocol.PLAINTEXT

    ssl_context: Optional[ssl.SSLContext] = None

    sasl_mechanism: SaslMechanism = SaslMechanism.PLAIN
    sasl_plain_username: Optional[str] = None
    sasl_plain_password: Optional[str] = None
    sasl_oauth_token_provider: Optional[str] = None
    model_config = ConfigDict(arbitrary_types_allowed=True, use_enum_values=True)

    @model_validator(mode="after")
    @classmethod
    def protocols_validation(cls, values):
        security_protocol = values.security_protocol

        if security_protocol == SecurityProtocol.PLAINTEXT:
            return values
        elif security_protocol == SecurityProtocol.SSL:
            if values.ssl_context is None:
                raise ValueError("`ssl_context` is required")
            return values
        elif security_protocol == SecurityProtocol.SASL_PLAINTEXT:
            if values.sasl_mechanism is SaslMechanism.OAUTHBEARER:
                # We don't perform a username and password check if OAUTHBEARER
                return values
            if (
                values.sasl_mechanism is SaslMechanism.PLAIN
                and values.sasl_plain_username is None
            ):
                raise ValueError(
                    "`sasl_plain_username` is required when using SASL_PLAIN"
                )
            if (
                values.sasl_mechanism is SaslMechanism.PLAIN
                and values.sasl_plain_password is None
            ):
                raise ValueError(
                    "`sasl_plain_password` is required when using SASL_PLAIN"
                )
            return values
        elif security_protocol == SecurityProtocol.SASL_SSL:
            if values.ssl_context is None:
                raise ValueError("`ssl_context` is required")
            if (
                values.sasl_mechanism is SaslMechanism.PLAIN
                and values.sasl_plain_username is None
            ):
                raise ValueError(
                    "`sasl_plain_username` is required when using SASL_PLAIN"
                )
            if (
                values.sasl_mechanism is SaslMechanism.PLAIN
                and values.sasl_plain_password is None
            ):
                raise ValueError(
                    "`sasl_plain_password` is required when using SASL_PLAIN"
                )
            return values



from typing import Any, Dict, Optional, Protocol

from kstreams import ConsumerRecord

from .types import Headers


class Deserializer(Protocol):
    """Protocol used by the Stream to deserialize.

    A Protocol is similar to other languages features like an interface or a trait.

    End users should provide their own class implementing this protocol.

    For example a `JsonDeserializer`

    ```python
    import json
    from kstreams import ConsumerRecord

    class JsonDeserializer:

        async def deserialize(
            self, consumer_record: ConsumerRecord, **kwargs
        ) -> ConsumerRecord:
            data = json.loads(consumer_record.value.decode())
            consumer_record.value = data
            return consumer_record
    ```
    """

    async def deserialize(
        self, consumer_record: ConsumerRecord, **kwargs
    ) -> ConsumerRecord:
        """
        Implement this method to deserialize the data received from the topic.
        """
        ...


class Serializer(Protocol):
    """Protocol used by the Stream to serialize.

    A Protocol is similar to other languages features like an interface or a trait.

    End users should provide their own class implementing this protocol.

    For example a `JsonSerializer`

    ```python
    from typing import Optional, Dict
    import json

    class JsonSerializer:

        async def serialize(
            self,
            payload: dict,
            headers: Optional[Dict[str, str]] = None,
            serializer_kwargs: Optional[Dict] = None,
        ) -> bytes:
            \"""Return UTF-8 encoded payload\"""
            value = json.dumps(payload)
            return value.encode()
    ```

    Notice that you don't need to inherit anything,
    you just have to comply with the Protocol.
    """

    async def serialize(
        self,
        payload: Any,
        headers: Optional[Headers] = None,
        serializer_kwargs: Optional[Dict] = None,
    ) -> bytes:
        """
        Implement this method to deserialize the data received from the topic.
        """
        ...



import logging
from typing import Callable, Optional, TypeVar

import aiokafka

logger = logging.getLogger(__name__)


class Consumer(aiokafka.AIOKafkaConsumer):
    def __init__(
        self,
        *args,
        key_deserializer: Optional[Callable] = None,
        **kwargs,
    ):
        self.config = kwargs

        if key_deserializer is None:

            def key_deserializer(key):
                if key is not None:
                    return key.decode()

        super().__init__(*args, key_deserializer=key_deserializer, **self.config)


class Producer(aiokafka.AIOKafkaProducer):
    def __init__(
        self,
        *args,
        key_serializer=None,
        **kwargs,
    ):
        self.config = kwargs

        if key_serializer is None:

            def key_serializer(key):
                if key is not None:
                    return key.encode("utf-8")

        super().__init__(*args, key_serializer=key_serializer, **self.config)


ConsumerType = TypeVar("ConsumerType", bound=Consumer)
ProducerType = TypeVar("ProducerType", bound=Producer)



import asyncio
import logging
from typing import DefaultDict, Dict, List, Optional, TypeVar

from prometheus_client import Gauge

from kstreams import TopicPartition
from kstreams.clients import ConsumerType
from kstreams.streams import Stream

logger = logging.getLogger(__name__)

PrometheusMonitorType = TypeVar("PrometheusMonitorType", bound="PrometheusMonitor")
MetricsType = Dict[TopicPartition, Dict[str, Optional[int]]]


class PrometheusMonitor:
    """
    Metrics monitor to keep track of Producers and Consumers.

     Attributes:
        metrics_scrape_time float: Amount of seconds that the monitor
            will wait until next scrape iteration
    """

    # Producer metrics
    MET_OFFSETS = Gauge(
        "topic_partition_offsets", "help producer offsets", ["topic", "partition"]
    )

    # Consumer metrics
    MET_COMMITTED = Gauge(
        "consumer_committed",
        "help consumer committed",
        ["topic", "partition", "consumer_group"],
    )
    MET_POSITION = Gauge(
        "consumer_position",
        "help consumer position",
        ["topic", "partition", "consumer_group"],
    )
    MET_HIGHWATER = Gauge(
        "consumer_highwater",
        "help consumer highwater",
        ["topic", "partition", "consumer_group"],
    )
    MET_LAG = Gauge(
        "consumer_lag",
        "help consumer lag calculated using the last commited offset",
        ["topic", "partition", "consumer_group"],
    )
    MET_POSITION_LAG = Gauge(
        "position_lag",
        "help consumer position lag calculated using the consumer position",
        ["topic", "partition", "consumer_group"],
    )

    def __init__(self, metrics_scrape_time: float = 3):
        self.metrics_scrape_time = metrics_scrape_time
        self.running = False
        self._producer = None
        self._streams: List[Stream] = []
        self._task: Optional[asyncio.Task] = None

    def start(self) -> None:
        logger.info("Starting Prometheus metrics...")
        self.running = True
        self._task = asyncio.create_task(self._metrics_task())

    async def stop(self) -> None:
        logger.info("Stopping Prometheus metrics...")
        self.running = False

        if self._task is not None:
            # we need to make sure that the task is `done`
            # to clean up properly
            while not self._task.done():
                await asyncio.sleep(0.1)

        self._clean_consumer_metrics()

    def add_topic_partition_offset(
        self, topic: str, partition: int, offset: int
    ) -> None:
        self.MET_OFFSETS.labels(topic=topic, partition=partition).set(offset)

    def _add_consumer_metrics(self, metrics_dict: MetricsType):
        for topic_partition, partitions_metadata in metrics_dict.items():
            group_id = partitions_metadata["group_id"]
            position = partitions_metadata["position"]
            committed = partitions_metadata["committed"]
            highwater = partitions_metadata["highwater"]
            lag = partitions_metadata["lag"]
            position_lag = partitions_metadata["position_lag"]

            self.MET_COMMITTED.labels(
                topic=topic_partition.topic,
                partition=topic_partition.partition,
                consumer_group=group_id,
            ).set(committed or 0)
            self.MET_POSITION.labels(
                topic=topic_partition.topic,
                partition=topic_partition.partition,
                consumer_group=group_id,
            ).set(position or -1)
            self.MET_HIGHWATER.labels(
                topic=topic_partition.topic,
                partition=topic_partition.partition,
                consumer_group=group_id,
            ).set(highwater or 0)
            self.MET_LAG.labels(
                topic=topic_partition.topic,
                partition=topic_partition.partition,
                consumer_group=group_id,
            ).set(lag or 0)
            self.MET_POSITION_LAG.labels(
                topic=topic_partition.topic,
                partition=topic_partition.partition,
                consumer_group=group_id,
            ).set(position_lag or 0)

    def _clean_consumer_metrics(self) -> None:
        """
        This method should be called when a rebalance takes place
        to clean all consumers metrics. When the rebalance finishes
        new metrics will be generated per consumer based on the
        consumer assigments
        """
        self.MET_LAG.clear()
        self.MET_POSITION_LAG.clear()
        self.MET_COMMITTED.clear()
        self.MET_POSITION.clear()
        self.MET_HIGHWATER.clear()

    def clean_stream_consumer_metrics(self, stream: Stream) -> None:
        if stream.consumer is not None:
            topic_partitions = stream.consumer.assignment()
            group_id = stream.consumer._group_id
            for topic_partition in topic_partitions:
                topic = topic_partition.topic
                partition = topic_partition.partition

                metrics_found = False
                for sample in self.MET_LAG.collect()[0].samples:
                    if {
                        "topic": topic,
                        "partition": str(partition),
                        "consumer_group": group_id,
                    } == sample.labels:
                        metrics_found = True

                if metrics_found:
                    self.MET_LAG.remove(topic, partition, group_id)
                    self.MET_POSITION_LAG.remove(topic, partition, group_id)
                    self.MET_COMMITTED.remove(topic, partition, group_id)
                    self.MET_POSITION.remove(topic, partition, group_id)
                    self.MET_HIGHWATER.remove(topic, partition, group_id)
                else:
                    logger.debug(f"Metrics for stream: {stream.name} not found")

    def add_producer(self, producer):
        self._producer = producer

    def add_streams(self, streams):
        self._streams = streams

    async def generate_consumer_metrics(self, consumer: ConsumerType):
        """
        Generate Consumer Metrics for Prometheus

        Format:
            {
                "topic-1": {
                    "1": (
                        [topic-1, partition-number, 'group-id-1'],
                        committed, position, highwater, lag, position_lag
                    )
                    "2": (
                        [topic-1, partition-number, 'group-id-1'],
                        committed, position, highwater, lag, position_lag
                    )
                },
                ...
                "topic-n": {
                    "1": (
                        [topic-n, partition-number, 'group-id-n'],
                        committed, position, highwater, lag, position_lag
                    )
                    "2": (
                        [topic-n, partition-number, 'group-id-n'],
                        committed, position, highwater, lag, position_lag
                    )
                }
            }
        """
        metrics: MetricsType = DefaultDict(dict)

        topic_partitions = consumer.assignment()

        for topic_partition in topic_partitions:
            committed = await consumer.committed(topic_partition) or 0
            position = await consumer.position(topic_partition)
            highwater = consumer.highwater(topic_partition)

            lag = position_lag = None
            if highwater:
                lag = highwater - committed
                position_lag = highwater - position

            metrics[topic_partition] = {
                "group_id": consumer._group_id,
                "committed": committed,
                "position": position,
                "highwater": highwater,
                "lag": lag,
                "position_lag": position_lag,
            }

        self._add_consumer_metrics(metrics)

    async def _metrics_task(self) -> None:
        """
        Asyncio Task that runs in `backgroud` to generate
        consumer metrics.

        When self.running is False the task will finish and it
        will be safe to stop consumers and producers.
        """
        while self.running:
            await asyncio.sleep(self.metrics_scrape_time)
            for stream in self._streams:
                if stream.consumer is not None:
                    try:
                        await self.generate_consumer_metrics(stream.consumer)
                    except RuntimeError:
                        logger.debug(
                            f"Metrics for stream {stream.name} can not be generated "
                            "probably because it has been removed"
                        )



from aiokafka.structs import ConsumerRecord, RecordMetadata, TopicPartition

from .clients import Consumer, ConsumerType, Producer, ProducerType
from .create import StreamEngine, create_engine
from .prometheus.monitor import PrometheusMonitor, PrometheusMonitorType
from .rebalance_listener import (
    ManualCommitRebalanceListener,
    MetricsRebalanceListener,
    RebalanceListener,
)
from .streams import Stream, stream
from .structs import TopicPartitionOffset
from .test_utils import TestStreamClient

__all__ = [
    "Consumer",
    "ConsumerType",
    "Producer",
    "ProducerType",
    "StreamEngine",
    "create_engine",
    "PrometheusMonitor",
    "PrometheusMonitorType",
    "MetricsRebalanceListener",
    "ManualCommitRebalanceListener",
    "RebalanceListener",
    "Stream",
    "stream",
    "ConsumerRecord",
    "RecordMetadata",
    "TestStreamClient",
    "TopicPartition",
    "TopicPartitionOffset",
]



import asyncio
import inspect
import logging
import sys
import typing
import uuid
from functools import update_wrapper

from aiokafka import errors

from kstreams import ConsumerRecord, TopicPartition
from kstreams.exceptions import BackendNotSet
from kstreams.structs import TopicPartitionOffset

from .backends.kafka import Kafka
from .clients import Consumer, ConsumerType
from .middleware import Middleware
from .rebalance_listener import RebalanceListener
from .serializers import Deserializer
from .streams_utils import UDFType, inspect_udf
from .types import StreamFunc

logger = logging.getLogger(__name__)


if sys.version_info < (3, 10):

    async def anext(async_gen: typing.AsyncGenerator):
        return await async_gen.__anext__()


class Stream:
    """
    Attributes:
        name str: Stream name
        topics List[str]: List of topics to consume
        backend kstreams.backends.Kafka: backend kstreams.backends.kafka.Kafka:
            Backend to connect. Default `Kafka`
        func Callable[["Stream"], Awaitable[Any]]: Coroutine fucntion or generator
            to be called when an event arrives
        config Dict[str, Any]: Stream configuration. Here all the
         [properties](https://aiokafka.readthedocs.io/en/stable/api.html#consumer-class)
            can be passed in the dictionary
        deserializer kstreams.serializers.Deserializer: Deserializer to be used
            when an event is consumed
        initial_offsets List[kstreams.TopicPartitionOffset]: List of
            TopicPartitionOffset that will `seek` the initial offsets to
        rebalance_listener kstreams.rebalance_listener.RebalanceListener: Listener
            callbacks when partition are assigned or revoked

    !!! Example
        ```python title="Usage"
        import aiorun
        from kstreams import create_engine, ConsumerRecord

        stream_engine = create_engine(title="my-stream-engine")


        # here you can add any other AIOKafkaConsumer config
        @stream_engine.stream("local--kstreams", group_id="de-my-partition")
        async def stream(cr: ConsumerRecord) -> None:
            print(f"Event consumed: headers: {cr.headers}, payload: {cr.value}")


        async def start():
            await stream_engine.start()
            await produce()


        async def shutdown(loop):
            await stream_engine.stop()


        if __name__ == "__main__":
            aiorun.run(
                start(),
                stop_on_unhandled_errors=True,
                shutdown_callback=shutdown
            )
        ```
    """

    def __init__(
        self,
        topics: typing.Union[typing.List[str], str],
        *,
        func: StreamFunc,
        backend: typing.Optional[Kafka] = None,
        consumer_class: typing.Type[ConsumerType] = Consumer,
        name: typing.Optional[str] = None,
        config: typing.Optional[typing.Dict] = None,
        deserializer: typing.Optional[Deserializer] = None,
        initial_offsets: typing.Optional[typing.List[TopicPartitionOffset]] = None,
        rebalance_listener: typing.Optional[RebalanceListener] = None,
        middlewares: typing.Optional[typing.List[Middleware]] = None,
    ) -> None:
        self.func = func
        self.backend = backend
        self.consumer_class = consumer_class
        self.consumer: typing.Optional[ConsumerType] = None
        self.config = config or {}
        self._consumer_task: typing.Optional[asyncio.Task] = None
        self.name = name or str(uuid.uuid4())
        self.deserializer = deserializer
        self.running = False
        self.initial_offsets = initial_offsets
        self.seeked_initial_offsets = False
        self.rebalance_listener = rebalance_listener
        self.udf_type = inspect_udf(func, Stream)
        self.middlewares = middlewares or []

        # aiokafka expects topic names as arguments, meaning that
        # can receive N topics -> N arguments,
        # so we always create a list and then we expand it with *topics
        self.topics = [topics] if isinstance(topics, str) else topics

    def _create_consumer(self) -> ConsumerType:
        if self.backend is None:
            raise BackendNotSet("A backend has not been set for this stream")
        config = {**self.backend.model_dump(), **self.config}
        return self.consumer_class(**config)

    async def stop(self) -> None:
        if not self.running:
            return None

        if self.consumer is not None:
            await self.consumer.stop()
            self.running = False

        if self._consumer_task is not None:
            self._consumer_task.cancel()

    async def _subscribe(self) -> None:
        # Always create a consumer on stream.start
        self.consumer = self._create_consumer()

        # add the chech tp avoid `mypy` complains
        if self.consumer is not None:
            await self.consumer.start()
            self.consumer.subscribe(
                topics=self.topics, listener=self.rebalance_listener
            )
            self.running = True

    async def commit(
        self, offsets: typing.Optional[typing.Dict[TopicPartition, int]] = None
    ):
        await self.consumer.commit(offsets=offsets)  # type: ignore

    async def getone(self) -> ConsumerRecord:
        consumer_record: ConsumerRecord = await self.consumer.getone()  # type: ignore

        # call deserializer if there is one regarless consumer_record.value
        # as the end user might want to do something extra with headers or metadata
        if self.deserializer is not None:
            return await self.deserializer.deserialize(consumer_record)

        return consumer_record

    async def getmany(
        self,
        partitions: typing.Optional[typing.List[TopicPartition]] = None,
        timeout_ms: int = 0,
        max_records: typing.Optional[int] = None,
    ) -> typing.Dict[TopicPartition, typing.List[ConsumerRecord]]:
        """
        Get a batch of events from the assigned TopicPartition.

        Prefetched events are returned in batches by topic-partition.
        If messages is not available in the prefetched buffer this method waits
        `timeout_ms` milliseconds.

        Attributes:
            partitions List[TopicPartition] | None: The partitions that need
                fetching message. If no one partition specified then all
                subscribed partitions will be used
            timeout_ms int | None: milliseconds spent waiting if
                data is not available in the buffer. If 0, returns immediately
                with any records that are available currently in the buffer,
                else returns empty. Must not be negative.
            max_records int | None: The amount of records to fetch.
                if `timeout_ms` was defined and reached and the fetched records
                has not reach `max_records` then returns immediately
                with any records that are available currently in the buffer

        Returns:
            Topic to list of records

        !!! Example
            ```python
            @stream_engine.stream(topic, ...)
            async def stream(stream: Stream):
                while True:
                    data = await stream.getmany(max_records=5)
                    print(data)
            ```
        """
        partitions = partitions or []
        return await self.consumer.getmany(  # type: ignore
            *partitions, timeout_ms=timeout_ms, max_records=max_records
        )

    async def start(self) -> None:
        if self.running:
            return None

        await self._subscribe()

        if self.udf_type == UDFType.NO_TYPING:
            # normal use case
            logging.warn(
                "Streams with `async for in` loop approach might be deprecated. "
                "Consider migrating to a typing approach."
            )

            func = self.func(self)
            # create an asyncio.Task with func
            self._consumer_task = asyncio.create_task(self.func_wrapper(func))
        else:
            # Typing cases
            if not inspect.isasyncgenfunction(self.func):
                # Is not an async_generator, then create an asyncio.Task with func
                self._consumer_task = asyncio.create_task(
                    self.func_wrapper_with_typing()
                )
        return None

    async def func_wrapper(self, func: typing.Awaitable) -> None:
        try:
            # await for the end user coroutine
            # we do this to show a better error message to the user
            # when the coroutine fails
            await func
        except Exception as e:
            logger.exception(f"CRASHED Stream!!! Task {self._consumer_task} \n\n {e}")

    async def func_wrapper_with_typing(self) -> None:
        while True:
            cr = await self.getone()
            await self.func(cr)

    def seek_to_initial_offsets(self):
        if not self.seeked_initial_offsets:
            assignments: typing.Set[TopicPartition] = self.consumer.assignment()
            if self.initial_offsets is not None:
                topicPartitionOffset: TopicPartitionOffset
                for topicPartitionOffset in self.initial_offsets:
                    tp = TopicPartition(
                        topic=topicPartitionOffset.topic,
                        partition=topicPartitionOffset.partition,
                    )
                    if tp in assignments:
                        self.consumer.seek(
                            partition=tp, offset=topicPartitionOffset.offset
                        )
                    else:
                        logger.warning(
                            "You are attempting to seek on an TopicPartitionOffset "
                            "that isn't in the consumer assignments. "
                            "The code will simply ignore the seek request "
                            f"on this partition. {tp} is not in the partition "
                            f"assignment. The partition assignment is {assignments}."
                        )
            self.seeked_initial_offsets = True

    async def __aenter__(self) -> "Stream":
        """
        Start the kafka Consumer and return an `async_gen` so it can be iterated

        !!! Example
            ```python title="Usage"
            @stream_engine.stream(topic, group_id=group_id, ...)
            async def stream(stream):
                yield cr.value


            # Iterate the stream:
            async with stream as stream_flow:
                async for value in stream_flow:
                    ...
            ```
        """
        logger.info("Starting async_gen Stream....")
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        logger.info("Stopping async_gen Stream....")
        await self.stop()

    def __aiter__(self):
        return self

    async def __anext__(self) -> ConsumerRecord:
        try:
            cr = await self.getone()

            if self.udf_type == UDFType.NO_TYPING:
                return cr
            return await self.func(cr)
        except errors.ConsumerStoppedError:
            raise StopAsyncIteration  # noqa: F821


class UdfHandler:
    def __init__(
        self,
        *,
        handler: StreamFunc,
        stream: Stream,
    ) -> None:
        self.handler = handler
        self.stream = stream
        self.udf_type = inspect_udf(self.handler, Stream)

    def get_udf_params(self, cr: ConsumerRecord) -> typing.Tuple:
        if self.udf_type == UDFType.CR_ONLY_TYPING:
            return (cr,)
        return (
            cr,
            self.stream,
        )

    async def __call__(self, cr: ConsumerRecord) -> typing.Any:
        """
        Call the coroutine `async def my_function(...)` defined by the end user
        in a proper way according to its parameters. The `handler` is the
        coroutine defined by the user.

        Use cases:
            1. UDFType.CR_ONLY_TYPING: Only ConsumerRecord with typing

            @stream_engine.stream(topic, name="my-stream")
                async def consume(cr: ConsumerRecord):
                    ...

            2. UDFType.ALL_TYPING: ConsumerRecord and Stream with typing.
                The order is important as they are arguments and not kwargs

            @stream_engine.stream(topic, name="my-stream")
                async def consume(cr: ConsumerRecord, stream: Stream):
                    ...

        """
        params = self.get_udf_params(cr)

        if inspect.isasyncgenfunction(self.handler):
            return await anext(self.handler(*params))
        return await self.handler(*params)


def stream(
    topics: typing.Union[typing.List[str], str],
    *,
    name: typing.Optional[str] = None,
    deserializer: typing.Optional[Deserializer] = None,
    initial_offsets: typing.Optional[typing.List[TopicPartitionOffset]] = None,
    rebalance_listener: typing.Optional[RebalanceListener] = None,
    middlewares: typing.Optional[typing.List[Middleware]] = None,
    **kwargs,
) -> typing.Callable[[StreamFunc], Stream]:
    def decorator(func: StreamFunc) -> Stream:
        s = Stream(
            topics=topics,
            func=func,
            name=name,
            deserializer=deserializer,
            initial_offsets=initial_offsets,
            rebalance_listener=rebalance_listener,
            middlewares=middlewares,
            config=kwargs,
        )
        update_wrapper(s, func)
        return s

    return decorator



import logging
import typing

from aiokafka import errors

from kstreams import ConsumerRecord, types

if typing.TYPE_CHECKING:
    from kstreams import Stream  #  pragma: no cover


logger = logging.getLogger(__name__)


class MiddlewareProtocol(typing.Protocol):
    def __init__(
        self,
        *,
        next_call: types.NextMiddlewareCall,
        send: types.Send,
        stream: "Stream",
        **kwargs: typing.Any,
    ) -> None:
        ...  #  pragma: no cover

    async def __call__(self, cr: ConsumerRecord) -> typing.Any:
        ...  #  pragma: no cover


class Middleware:
    def __init__(
        self, middleware: typing.Type[MiddlewareProtocol], **kwargs: typing.Any
    ) -> None:
        self.middleware = middleware
        self.kwargs = kwargs

    def __iter__(self) -> typing.Iterator:
        return iter((self.middleware, self.kwargs))


class BaseMiddleware:
    def __init__(
        self,
        *,
        next_call: types.NextMiddlewareCall,
        send: types.Send,
        stream: "Stream",
    ) -> None:
        self.next_call = next_call
        self.send = send
        self.stream = stream

    async def __call__(self, cr: ConsumerRecord) -> typing.Any:
        raise NotImplementedError


class ExceptionMiddleware(BaseMiddleware):
    async def __call__(self, cr: ConsumerRecord) -> typing.Any:
        try:
            return await self.next_call(cr)
        except errors.ConsumerStoppedError as exc:
            logger.exception(
                f"Stream consuming from topics {self.stream.topics} has stopped!!! \n\n"
            )
            await self.stream.stop()
            raise exc
        except Exception as exc:
            logger.exception(
                f"Stream consuming from topics {self.stream.topics} CRASHED!!! \n\n "
            )
            await self.stream.stop()
            raise exc



import asyncio
import logging
from typing import Set

from aiokafka.abc import ConsumerRebalanceListener

from kstreams import TopicPartition

logger = logging.getLogger(__name__)


# Can not use a Protocol here because aiokafka forces to have a concrete instance
# that inherits from ConsumerRebalanceListener, if we use a protocol we will
# have to force the end users to import the class and inherit from it,
# then we will mix protocols and inheritance
class RebalanceListener(ConsumerRebalanceListener):
    """
    A callback interface that the user can implement to trigger custom actions
    when the set of partitions are assigned or revoked to the `Stream`.

    !!! Example
        ```python
        from kstreams import RebalanceListener, TopicPartition
        from .resource import stream_engine


        class MyRebalanceListener(RebalanceListener):

            async def on_partitions_revoked(
                self, revoked: Set[TopicPartition]
            ) -> None:
                # Do something with the revoked partitions
                # or with the Stream
                print(self.stream)

            async def on_partitions_assigned(
                self, assigned: Set[TopicPartition]
            ) -> None:
                # Do something with the assigned partitions
                # or with the Stream
                print(self.stream)


        @stream_engine.stream(topic, rebalance_listener=MyRebalanceListener())
        async def my_stream(stream: Stream):
            async for event in stream:
                ...
        ```
    """

    def __init__(self) -> None:
        self.stream = None
        # engine added so it can react on rebalance events
        self.engine = None

    async def on_partitions_revoked(self, revoked: Set[TopicPartition]) -> None:
        """
        Coroutine to be called *before* a rebalance operation starts and
        *after* the consumer stops fetching data.

        If you are using manual commit you have to commit all consumed offsets
        here, to avoid duplicate message delivery after rebalance is finished.

        Use cases:
            - cleanup or custom state save on the start of a rebalance operation
            - saving offsets in a custom store

        Attributes:
            revoked Set[TopicPartitions]: Partitions that were assigned
                to the consumer on the last rebalance

        !!! note
            The `Stream` is available using `self.stream`
        """
        ...  # pragma: no cover

    async def on_partitions_assigned(self, assigned: Set[TopicPartition]) -> None:
        """
        Coroutine to be called *after* partition re-assignment completes
        and *before* the consumer starts fetching data again.

        It is guaranteed that all the processes in a consumer group will
        execute their `on_partitions_revoked` callback before any instance
        executes its `on_partitions_assigned` callback.

        Use cases:
            - Load a state or cache warmup on completion of a successful
            partition re-assignment.

        Attributes:
            assigned Set[TopicPartition]: Partitions assigned to the
                consumer (may include partitions that were previously assigned)

        !!! note
            The `Stream` is available using `self.stream`
        """
        ...  # pragma: no cover


class MetricsRebalanceListener(RebalanceListener):
    async def on_partitions_revoked(self, revoked: Set[TopicPartition]) -> None:
        """
        Coroutine to be called *before* a rebalance operation starts and
        *after* the consumer stops fetching data.

        This will method will clean up the `Prometheus` metrics

        Attributes:
            revoked Set[TopicPartitions]: Partitions that were assigned
                to the consumer on the last rebalance
        """
        # lock all asyncio Tasks so no new metrics will be added to the Monitor
        if revoked and self.engine is not None:
            async with asyncio.Lock():
                await self.engine.monitor.stop()

    async def on_partitions_assigned(self, assigned: Set[TopicPartition]) -> None:
        """
        Coroutine to be called *after* partition re-assignment completes
        and *before* the consumer starts fetching data again.

        This method will start the `Prometheus` metrics

        Attributes:
            assigned Set[TopicPartition]: Partitions assigned to the
                consumer (may include partitions that were previously assigned)
        """
        # lock all asyncio Tasks so no new metrics will be added to the Monitor
        if assigned and self.engine is not None:
            async with asyncio.Lock():
                self.engine.monitor.start()

                stream = self.stream
                if stream is not None:
                    stream.seek_to_initial_offsets()


class ManualCommitRebalanceListener(MetricsRebalanceListener):
    async def on_partitions_revoked(self, revoked: Set[TopicPartition]) -> None:
        """
        Coroutine to be called *before* a rebalance operation starts and
        *after* the consumer stops fetching data.

        If manual commit is enabled, `commit` is called before the consumers
        partitions are revoked to prevent the error `CommitFailedError`
        and duplicate message delivery after a rebalance.

        Attributes:
            revoked Set[TopicPartitions]: Partitions that were assigned
                to the consumer on the last rebalance
        """
        if (
            revoked
            and self.stream is not None
            and not self.stream.consumer._enable_auto_commit
        ):
            logger.info(
                f"Manual commit enabled for stream {self.stream}. "
                "Performing `commit` before revoking partitions"
            )
            async with asyncio.Lock():
                await self.stream.commit()

            await super().on_partitions_revoked(revoked=revoked)



class StreamException(Exception):
    """Used to catch all kstream exceptions."""


class DuplicateStreamException(StreamException):
    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name} {self.name}"

    def __str__(self) -> str:
        msg = (
            f"Duplicate Stream found with name {self.name} "
            "Make sure that the name is unique"
        )

        return msg


class EngineNotStartedException(StreamException):
    def __str__(self) -> str:
        return "Engine has not been started. Try with `await engine.start()`"


class BackendNotSet(StreamException):
    ...



import inspect
import logging
import typing

from aiokafka.structs import RecordMetadata

from kstreams.structs import TopicPartitionOffset

from .backends.kafka import Kafka
from .clients import ConsumerType, ProducerType
from .exceptions import DuplicateStreamException, EngineNotStartedException
from .middleware import ExceptionMiddleware, Middleware
from .prometheus.monitor import PrometheusMonitor
from .rebalance_listener import MetricsRebalanceListener, RebalanceListener
from .serializers import Deserializer, Serializer
from .streams import Stream, StreamFunc, UdfHandler
from .streams import stream as stream_func
from .streams_utils import UDFType, inspect_udf
from .types import Headers, NextMiddlewareCall
from .utils import encode_headers

logger = logging.getLogger(__name__)


class StreamEngine:
    """
    Attributes:
        backend kstreams.backends.Kafka: Backend to connect. Default `Kafka`
        consumer_class kstreams.Consumer: The consumer class to use when
            instanciate a consumer. Default kstreams.Consumer
        producer_class kstreams.Producer: The producer class to use when
            instanciate the producer. Default kstreams.Producer
        monitor kstreams.PrometheusMonitor: Prometheus monitor that holds
            the [metrics](https://kpn.github.io/kstreams/metrics/)
        title str | None: Engine name
        serializer kstreams.serializers.Serializer | None: Serializer to
            use when an event is produced.
        deserializer kstreams.serializers.Deserializer | None: Deserializer
            to be used when an event is consumed.
            If provided it will be used in all Streams instances as a general one.
            To override it per Stream, you can provide one per Stream

    !!! Example
        ```python title="Usage"
        import kstreams

        stream_engine = kstreams.create_engine(
            title="my-stream-engine"
        )

        @kstreams.stream("local--hello-world", group_id="example-group")
        async def consume(stream: kstreams.ConsumerRecord) -> None:
            print(f"showing bytes: {cr.value}")


        await stream_engine.start()
        ```
    """

    def __init__(
        self,
        *,
        backend: Kafka,
        consumer_class: typing.Type[ConsumerType],
        producer_class: typing.Type[ProducerType],
        monitor: PrometheusMonitor,
        title: typing.Optional[str] = None,
        deserializer: typing.Optional[Deserializer] = None,
        serializer: typing.Optional[Serializer] = None,
    ) -> None:
        self.title = title
        self.backend = backend
        self.consumer_class = consumer_class
        self.producer_class = producer_class
        self.deserializer = deserializer
        self.serializer = serializer
        self.monitor = monitor
        self._producer: typing.Optional[typing.Type[ProducerType]] = None
        self._streams: typing.List[Stream] = []

    async def send(
        self,
        topic: str,
        value: typing.Any = None,
        key: typing.Any = None,
        partition: typing.Optional[int] = None,
        timestamp_ms: typing.Optional[int] = None,
        headers: typing.Optional[Headers] = None,
        serializer: typing.Optional[Serializer] = None,
        serializer_kwargs: typing.Optional[typing.Dict] = None,
    ):
        """
        Attributes:
            topic str: Topic name to send the event to
            value Any: Event value
            key str | None: Event key
            partition int | None: Topic partition
            timestamp_ms int | None: Event timestamp in miliseconds
            headers Dict[str, str] | None: Event headers
            serializer kstreams.serializers.Serializer | None: Serializer to
                encode the event
            serializer_kwargs Dict[str, Any] | None: Serializer kwargs
        """
        if self._producer is None:
            raise EngineNotStartedException()

        serializer = serializer or self.serializer

        # serialize only when value and serializer are present
        if value is not None and serializer is not None:
            value = await serializer.serialize(
                value, headers=headers, serializer_kwargs=serializer_kwargs
            )

        encoded_headers = None
        if headers is not None:
            encoded_headers = encode_headers(headers)

        fut = await self._producer.send(
            topic,
            value=value,
            key=key,
            partition=partition,
            timestamp_ms=timestamp_ms,
            headers=encoded_headers,
        )
        metadata: RecordMetadata = await fut
        self.monitor.add_topic_partition_offset(
            topic, metadata.partition, metadata.offset
        )

        return metadata

    async def start(self) -> None:
        await self.start_producer()
        await self.start_streams()

        # add the producer and streams to the Monitor
        self.monitor.add_producer(self._producer)
        self.monitor.add_streams(self._streams)
        self.monitor.start()

    async def stop(self) -> None:
        await self.monitor.stop()
        await self.stop_producer()
        await self.stop_streams()

    async def stop_producer(self):
        logger.info("Waiting Producer to STOP....")
        if self._producer is not None:
            await self._producer.stop()

    async def start_producer(self, **kwargs) -> None:
        if self.producer_class is None:
            return None
        config = {**self.backend.model_dump(), **kwargs}
        self._producer = self.producer_class(**config)
        if self._producer is None:
            return None
        await self._producer.start()

    async def start_streams(self) -> None:
        # Only start the Streams that are not async_generators
        streams = [
            stream
            for stream in self._streams
            if not inspect.isasyncgenfunction(stream.func)
        ]
        for stream in streams:
            await stream.start()

    async def stop_streams(self) -> None:
        logger.info("Waiting for Streams to STOP....")
        for stream in self._streams:
            await stream.stop()

    async def clean_streams(self):
        await self.stop_streams()
        self._streams = []

    def exist_stream(self, name: str) -> bool:
        stream = self.get_stream(name)
        return True if stream is not None else False

    def get_stream(self, name: str) -> typing.Optional[Stream]:
        stream = next((stream for stream in self._streams if stream.name == name), None)

        return stream

    def add_stream(self, stream: Stream) -> None:
        if self.exist_stream(stream.name):
            raise DuplicateStreamException(name=stream.name)
        stream.backend = self.backend
        if stream.deserializer is None:
            stream.deserializer = self.deserializer
        self._streams.append(stream)

        if stream.rebalance_listener is None:
            # set the stream to the listener to it will be available
            # when the callbacks are called
            stream.rebalance_listener = MetricsRebalanceListener()

        stream.rebalance_listener.stream = stream  # type: ignore
        stream.rebalance_listener.engine = self  # type: ignore

        udf_type = inspect_udf(stream.func, Stream)
        if udf_type != UDFType.NO_TYPING:
            stream.func = self.build_stream_middleware_stack(stream)

    def build_stream_middleware_stack(self, stream: Stream) -> NextMiddlewareCall:
        udf_handler = UdfHandler(handler=stream.func, stream=stream)
        stream.middlewares = [Middleware(ExceptionMiddleware)] + stream.middlewares

        next_call = udf_handler
        for middleware, options in reversed(stream.middlewares):
            next_call = middleware(
                next_call=next_call, send=self.send, stream=stream, **options
            )
        return next_call

    async def remove_stream(self, stream: Stream) -> None:
        self._streams.remove(stream)
        await stream.stop()
        self.monitor.clean_stream_consumer_metrics(stream)

    def stream(
        self,
        topics: typing.Union[typing.List[str], str],
        *,
        name: typing.Optional[str] = None,
        deserializer: typing.Optional[Deserializer] = None,
        initial_offsets: typing.Optional[typing.List[TopicPartitionOffset]] = None,
        rebalance_listener: typing.Optional[RebalanceListener] = None,
        middlewares: typing.Optional[typing.List[Middleware]] = None,
        **kwargs,
    ) -> typing.Callable[[StreamFunc], Stream]:
        def decorator(func: StreamFunc) -> Stream:
            stream_from_func = stream_func(
                topics,
                name=name,
                deserializer=deserializer,
                initial_offsets=initial_offsets,
                rebalance_listener=rebalance_listener,
                middlewares=middlewares,
                **kwargs,
            )(func)
            self.add_stream(stream_from_func)

            return stream_from_func

        return decorator
